// copyright ############################### #
// This file is part of the Xcoll Package.   #
// Copyright (c) CERN, 2023.                 #
// ######################################### #

#ifndef XCOLL_STATES_H
#define XCOLL_STATES_H

#define  XC_LOST_ON_EVEREST         -333
#define  XC_LOST_ON_EVEREST_CRYSTAL -334
#define  XC_LOST_ON_FLUKA           -335
#define  XC_LOST_ON_GEANT4          -336
#define  XC_LOST_ON_ABSORBER        -340

#define  XC_ERR_INVALID_TRACK       -390
#define  XC_ERR_NOT_IMPLEMENTED     -391
#define  XC_ERR_INVALID_XOFIELD     -392
#define  XC_ERR                     -399
    
#endif /* XCOLL_STATES_H */
